package ij2x.plugin.filter;

import ij.CompositeImage;
import ij.IJ;
import ij.ImagePlus;
import ij.ImageStack;
import ij.Macro;
import ij.Undo;
import ij.WindowManager;
import ij.gui.GenericDialog;
import ij.gui.Roi;
import ij.gui.Toolbar;
import ij.measure.Calibration;
import ij.plugin.PlugIn;
import ij.plugin.Resizer;
import ij.process.ImageProcessor;
import ij.process.StackProcessor;
import ij.util.Tools;
import java.awt.Color;
import java.awt.Rectangle;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.util.Vector;

import javax.swing.text.Document;
import javax.swing.JTextField;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;
import javax.swing.JFormattedTextField;
import javax.swing.JComponent;
import javax.swing.JSpinner;

/** This plugin implements the Image/Scale command. */
public class ScaleAll implements PlugIn, DocumentListener, FocusListener {
        private ImagePlus imp;
        private static String xstr = "0.5";
        private static String ystr = "0.5";
        private static double xscale=0.5;
        private static double yscale=0.5;
        private String zstr = "1.0";
        private static int newWidth, newHeight;
        private int newDepth;
    private static boolean averageWhenDownsizing = true;
        private static boolean newWindow = true;
        private static int interpolationMethod = ImageProcessor.BILINEAR;
        private String[] methods = ImageProcessor.getInterpolationMethods();
        private static boolean fillWithBackground;
        private static boolean processStack = true;
        private double zscale;
        private String title = "Untitled";
        private Vector fields;
        private double bgValue;
        private boolean constainAspectRatio = true;
        private JTextField xField, yField, zField, widthField, heightField, depthField, titleField;
        private JSpinner xspinner, yspinner;
        private Vector<JComponent> spinnerFields=new Vector<JComponent>(), spinners=new Vector<JComponent>();
        private Rectangle r;
        private Object fieldWithFocus;
        private int oldDepth;
        static boolean isStack;

        public void run(String arg) {
                imp = IJ.getImage();
                Roi roi = imp.getRoi();
                if (roi!=null && !roi.isArea())
                        imp.killRoi(); // ignore any line selection
                ImageProcessor ip = imp.getProcessor();
                if (!showDialog(ip))
                        return;
                if (newDepth>0 && newDepth!=imp.getStackSize()) {
                        newWindow = true;
                        processStack = true;
                }
                if (ip.getWidth()>1 && ip.getHeight()>1)
                        ip.setInterpolationMethod(interpolationMethod);
                else
                        ip.setInterpolationMethod(ImageProcessor.NONE);
                ip.setBackgroundValue(bgValue);
                imp.startTiming();
                try {
                        if (newWindow && imp.getStackSize()>1 && processStack)
                                createNewStack(imp, ip);
                        else
                                scale(ip);
                }
                catch(OutOfMemoryError o) {
                        IJ.outOfMemory("Scale");
                }
                IJ.showProgress(1.0);
        }

        void createNewStack(ImagePlus imp, ImageProcessor ip) {
                int nSlices = imp.getStackSize();
                int w=imp.getWidth(), h=imp.getHeight();
                ImagePlus imp2 = imp.createImagePlus();
                if (newWidth!=w || newHeight!=h) {
                        Rectangle r = ip.getRoi();
                        boolean crop = r.width!=imp.getWidth() || r.height!=imp.getHeight();
                        ImageStack stack1 = imp.getStack();
                        ImageStack stack2 = new ImageStack(newWidth, newHeight);
                        ImageProcessor ip1, ip2;
                        int method = interpolationMethod;
                        if (w==1 || h==1)
                                method = ImageProcessor.NONE;
                        for (int i=1; i<=nSlices; i++) {
                                IJ.showStatus("Scale: " + i + "/" + nSlices);
                                ip1 = stack1.getProcessor(i);
                                String label = stack1.getSliceLabel(i);
                                if (crop) {
                                        ip1.setRoi(r);
                                        ip1 = ip1.crop();
                                }
                                ip1.setInterpolationMethod(method);
                                ip2 = ip1.resize(newWidth, newHeight, averageWhenDownsizing);
                                if (ip2!=null)
                                        stack2.addSlice(label, ip2);
                                IJ.showProgress(i, nSlices);
                        }
                        imp2.setStack(title, stack2);
                        Calibration cal = imp2.getCalibration();
                        if (cal.scaled()) {
                                cal.pixelWidth *= 1.0/xscale;
                                cal.pixelHeight *= 1.0/yscale;
                        }
                        IJ.showProgress(1.0);
                } else
                        imp2.setStack(title, imp.getStack());
                int[] dim = imp.getDimensions();
                imp2.setDimensions(dim[2], dim[3], dim[4]);
                if (imp.isComposite()) {
                        imp2 = new CompositeImage(imp2, ((CompositeImage)imp).getMode());
                        ((CompositeImage)imp2).copyLuts(imp);
                }
                if (imp.isHyperStack())
                        imp2.setOpenAsHyperStack(true);
                if (newDepth>0 && newDepth!=oldDepth)
                        imp2 = (new Resizer()).zScale(imp2, newDepth, interpolationMethod);
                if (imp2!=null) {
                        imp2.show();
                        imp2.changes = true;
                }
        }

        void scale(ImageProcessor ip) {
                if (newWindow) {
                        Rectangle r = ip.getRoi();
                        ImagePlus imp2 = imp.createImagePlus();
                        imp2.setProcessor(title, ip.resize(newWidth, newHeight, averageWhenDownsizing));
                        Calibration cal = imp2.getCalibration();
                        if (cal.scaled()) {
                                cal.pixelWidth *= 1.0/xscale;
                                cal.pixelHeight *= 1.0/yscale;
                        }
                        imp2.show();
                        imp.trimProcessor();
                        imp2.trimProcessor();
                        imp2.changes = true;
                } else {
                        if (processStack && imp.getStackSize()>1) {
                                Undo.reset();
                                StackProcessor sp = new StackProcessor(imp.getStack(), ip);
                                sp.scale(xscale, yscale, bgValue);
                        } else {
                                ip.snapshot();
                                Undo.setup(Undo.FILTER, imp);
                                ip.setSnapshotCopyMode(true);
                                ip.scale(xscale, yscale);
                                ip.setSnapshotCopyMode(false);
                        }
                        imp.killRoi();
                        imp.updateAndDraw();
                        imp.changes = true;
                }
        }

        boolean showDialog(ImageProcessor ip) {
                String macroOptions = Macro.getOptions();
                if (macroOptions!=null) {
                        if (macroOptions.indexOf(" interpolate")!=-1)
                                macroOptions.replaceAll(" interpolate", " interpolation=Bilinear");
                        else if (macroOptions.indexOf(" interpolation=")==-1)
                                macroOptions = macroOptions+" interpolation=None";
                        Macro.setOptions(macroOptions);
                }
                int bitDepth = imp.getBitDepth();
                int stackSize = imp.getStackSize();
                isStack = stackSize>1;
                oldDepth = stackSize;
                if (isStack) {
                        xstr = "1.0";
                        ystr = "1.0";
                        zstr = "1.0";
                }
                r = ip.getRoi();
                int width = newWidth;
                if (width==0) width = r.width;
                int height = (int)((double)width*r.height/r.width);
                zscale = 1.0;
                if (xscale!=0.0 && yscale!=0.0) {
                        width = (int)(r.width*xscale);
                        height = (int)(r.height*yscale);
                } else {
                        xstr = "-";
                        ystr = "-";
                }
                GenericDialog gd = new GenericDialog("Scale");
                String scaleXLabel="X Scale(0.05-25):";
                String scaleYLabel="Y Scale(0.05-25):";
                double min=0.05, max=25.0, value=0.5;
                if(xscale < min || yscale < min){xscale=min;yscale=min;}
                gd.addSpinner(scaleXLabel, min, max, xscale, 0, 0);
                gd.addSpinner(scaleYLabel, min, max, yscale, 0, 0);
                spinnerFields = gd.getSpinnerFields();
                int spinnerSize = spinnerFields.size();
                JFormattedTextField[] formatTextField = new JFormattedTextField[spinnerSize];
                spinners = gd.getSpinners();
                for (int i=0; i<spinnerSize; i++) {
                    formatTextField[i]=(JFormattedTextField)spinnerFields.elementAt(i);
                    formatTextField[i].addFocusListener(this);
                    formatTextField[i].getDocument().addDocumentListener(this);
                }
                xField = formatTextField[0];
                yField = formatTextField[1];
                xspinner = (JSpinner)spinners.elementAt(0);
                yspinner = (JSpinner)spinners.elementAt(1);
                if (isStack)
                        gd.addStringField("Z Scale:", zstr);
                gd.setInsets(5, 0, 5);
                gd.addStringField("Width (pixels):", ""+width);
                gd.addStringField("Height (pixels):", ""+height);
                if (isStack) {
                        String label = "Depth (images):";
                        if (imp.isHyperStack()) {
                                int slices = imp.getNSlices();
                                int frames = imp.getNFrames();
                                if (slices==1&&frames>1) {
                                        label = "Depth (frames):";
                                        oldDepth = frames;
                                } else {
                                        label = "Depth (slices):";
                                        oldDepth = slices;
                                }
                        }
                        gd.addStringField(label, ""+oldDepth);
                }
                gd.addChoice("Interpolation:", methods, methods[interpolationMethod]);
                if (bitDepth==8 || bitDepth==24)
                        gd.addCheckbox("Fill with background color", fillWithBackground);
                gd.addCheckbox("Average when downsizing", averageWhenDownsizing);
                if (isStack)
                        gd.addCheckbox("Process entire stack", processStack);
                gd.addCheckbox("Create new window", newWindow);
                title = WindowManager.getUniqueName(imp.getTitle());
                gd.setInsets(0, 0, 0);
                gd.addStringField("Title:", title, 12);
                fields = gd.getStringFields();
                JTextField[] allField = new JTextField[fields.size()];
                for (int i=0; i<fields.size(); i++) {
                        allField[i]=(JTextField)fields.elementAt(i);
                        allField[i].getDocument().addDocumentListener(this);
                        allField[i].addFocusListener(this);

                }
                if (!isStack) {
                    widthField  = allField[0];
                    heightField = allField[1];
                    titleField  = allField[2];
               } else {

                    zField = allField[0];
                    widthField  = allField[1];
                    heightField = allField[2];
                    depthField  = allField[3];
                    titleField  = allField[4];
                }
                xField.setColumns(allField[0].getColumns()-1);
                yField.setColumns(allField[0].getColumns()-1);
                fieldWithFocus = xField;
                gd.showDialog();
                if (gd.wasCanceled())
                        return false;
                    Double dx = (Double)xspinner.getValue();
                    Double dy = (Double)yspinner.getValue();
                xscale = (Double)xspinner.getValue();
                yscale = (Double)yspinner.getValue();
                yscale = xscale;
                if (isStack) {
                        zstr = zField.getText();
                        zscale = Tools.parseDouble(ystr, 0.0);
                }
                String wstr = widthField.getText();
                newWidth = (int)Tools.parseDouble(wstr, 0);
                newHeight = (int)Tools.parseDouble(heightField.getText(), 0);
                if (newHeight!=0 && (wstr.equals("-") || wstr.equals("0")))
                                newWidth= (int)(newHeight*(double)r.width/r.height);
                if (newWidth==0 || newHeight==0) {
                        IJ.error("Scaler", "Width or height is 0");
                        return false;
                }
                if (xscale>25.0) xscale = 25.0;
                if (yscale>25.0) yscale = 25.0;
                if (xscale>0.0 && yscale>0.0) {
                        newWidth = (int)(r.width*xscale);
                        newHeight = (int)(r.height*yscale);
                }
                if (isStack)
                        newDepth = (int)Tools.parseDouble(depthField.getText(), 0);//gd.getNextString(), 0);
                interpolationMethod = gd.getNextChoiceIndex();
                if (bitDepth==8 || bitDepth==24)
                        fillWithBackground = gd.getNextBoolean();
                averageWhenDownsizing = gd.getNextBoolean();
                if (isStack)
                        processStack = gd.getNextBoolean();
                newWindow = gd.getNextBoolean();
                if (xscale==0.0) {
                        xscale = (double)newWidth/r.width;
                        yscale = (double)newHeight/r.height;
                }
                title = titleField.getText();//)gd.getNextString();

                if (fillWithBackground) {
                        Color bgc = Toolbar.getBackgroundColor();
                        if (bitDepth==8)
                                bgValue = ip.getBestIndex(bgc);
                        else if (bitDepth==24)
                                bgValue = bgc.getRGB();
                } else
                        bgValue = 0.0;
                return true;
        }

        public void textValueChanged(DocumentEvent e) {
                Document source = e.getDocument();
                double newXScale = xscale;
                double newYScale = yscale;
                double newZScale = zscale;
                if (source==xField.getDocument() && fieldWithFocus==xField) {
                        String newXText = xField.getText();
                        newXScale = Tools.parseDouble(newXText,0);
                        if (newXScale==0) return;
                        if (newXScale!=xscale) {
                                int newWidth = (int)(newXScale*r.width);
                                widthField.setText(""+newWidth);
                                if (constainAspectRatio) {
                                        yField.setText(newXText);
                                        int newHeight = (int)(newXScale*r.height);
                                        heightField.setText(""+newHeight);
                                }
                        }
                } else if (source==yField.getDocument() && fieldWithFocus==yField) {
                        String newYText = yField.getText();
                        newYScale = Tools.parseDouble(newYText,0);
                        if (newYScale==0) return;
                        if (newYScale!=yscale) {
                                int newHeight = (int)(newYScale*r.height);
                                heightField.setText(""+newHeight);
                        }
                } else if (isStack && source==zField.getDocument() && fieldWithFocus==zField) {
                        String newZText = zField.getText();
                        newZScale = Tools.parseDouble(newZText,0);
                        if (newZScale==0) return;
                        if (newZScale!=zscale) {
                                int newDepth= (int)(newZScale*imp.getStackSize());
                                depthField.setText(""+newDepth);
                        }
                } else if (isStack && source==widthField.getDocument() && fieldWithFocus==widthField) {
                        int newWidth = (int)Tools.parseDouble(widthField.getText(), 0.0);
                        if (newWidth!=0) {
                                int newHeight = (int)(newWidth*(double)r.height/r.width);
                                heightField.setText(""+newHeight);
                                xField.setText("-");
                                yField.setText("-");
                                newXScale = 0.0;
                                newYScale = 0.0;
                        }
       } else if (isStack && source==depthField.getDocument() && fieldWithFocus==depthField) {
            int newDepth = (int)Tools.parseDouble(depthField.getText(), 0.0);
            if (newDepth!=0) {
                zField.setText("-");
                newZScale = 0.0;
            }
        }
                xscale = newXScale;
                yscale = newYScale;
                zscale = newZScale;
        }
        public void insertUpdate(DocumentEvent e){
            textValueChanged(e);
        }
        public void removeUpdate(DocumentEvent e){
            textValueChanged(e);
        }
        public void changedUpdate(DocumentEvent e){
        }

        public void focusGained(FocusEvent e) {
                fieldWithFocus = e.getSource();
                if (fieldWithFocus==widthField)
                        constainAspectRatio = true;
                else if (fieldWithFocus==yField)
                        constainAspectRatio = false;
        }

        public void focusLost(FocusEvent e) {}

}
